# er1 python interface
#
# InitializeRobot(PORT) - (port # from Device Manager control panel)
#
# Go(cmpersec, degpersec=0.0) - sets translational, rotational velocities
#
# Wh() - returns a tuple of (x,y,th) odometry
#
# PowerOff() - turns off power to the motors Go(...) turns it back on
#              note there is a difference between Go(0) and PowerOff():
#              Go(0) will actively hold a velocity of 0 (e.g., on hills)
#              PowerOff() lets the wheels float (e.g., roll down a hill)
#
# CloseRobot() - powers off and closes the serial port

#
# 10/1/2005
#
# Windows instructions for using this python ER1 driver
#
# make sure python is installed
#
# install Mark Hammond's Win 32 python extensions for Windows:
# http://starship.python.net/crew/mhammond/win32/Downloads.html
#
# you will know this has worked if you can type
# > import serial
# at the python prompt
#
# when you plug in the ER1 drive system (the RCM or Robot
# Control Module) for the first time, it will look for
# drivers for the device. 
#
# If you have the ER1 software CD, I would recommend _not_
# installing its software. Instead, copy the folder named
# "USB Robot Control Module" somewhere on your hard drive.
# When the system asks you to look for drivers, override
# the automatic search and then direct it to that folder.
# You will need to do this twice: once for the RCM and once
# for the USB-serial bridge. This will make the RCM look like an
# ordinary serial port device, listed in the device manager
# under "Ports" as COM17 or COM 5 or whatever...
#
# this COM number is the port number you'll need to pass into 
# InitializeRobot(PORT)
#
# if you do not have the ER1 CD, you should be able to download
# the drivers for free from www.ftdichip.com, the company that
# makes the communications chip that the RCM uses
#
# if it turns out to be legal to provide these drivers ourselves,
# they'll be included from this download location
#
# start python and type 
# > from er1 import *
# or
# > import er1    (and then use the er1. prefix throughout)
#
# to make the initial connection, run
# > InitializeRobot(PORT) # replace PORT with your COM port number
#
# if it succeeds, try
# > Go(5)
# the motors should move forward if motor2 is attached to the
# right motor and motor1 is attached to the left motor
#
# > Go(0)
# the motors should stop
#
# > Wh()
# the current odometry should be returned
# Note: the odometry depends on the robot's kinematics
#       they are implemented here for the default configuration
# this can be accessed as the global value gpos, as well
#
# > PowerOff()
# the motors should float
#
# > CloseRobot()
# closes the connections to the robot
# the power will go on for a moment as the velocity is set to 0
#
#
# It's good to turn off the robot power after this.
# As far as we can tell, there is still something
# happening even after CloseRobot() and turning
# off the power (or toggling it) stops that.
#
# Zach Dodds dodds@cs.hmc.edu

# things to be done
#   er1 simulator
#   check for serial connection more robustly

# 11/25/05 - Jeremy Stolarz, jstolarz@cmu.edu
# Made a small change to serial port init code so that it would convert the
# PORT value to int before trying to open the port for Windows

import serial
import time
import math
import sys

DEBUG = 0
NOSEND = 0

# Kinematic constants
WHEEL_CIRCUMFERENCE = math.pi * 10    # Evolution's wheels are 10 cm in diameter
COUNTS_PER_REVOLUTION = 256*210.0     # Empirically determined for the standard ER1
DELTA = 19.0                          # 1/2 the distance bewteen the wheels = ~19cm

# stuff taken from the C++ interface
UPDATE                  = 0x1A
SET_MOTOR_COMMAND       = 0x77
RESET                   = 0x39
SET_LIMIT_SWITCH_MODE   = 0x80
SET_PROFILE_MODE        = 0xA0
SET_ACCELERATION        = 0x90
SET_DECELERATION        = 0x91
SET_JERK                = 0x13
SET_VELOCITY            = 0x11
GET_COMMANDED_POSITION  = 0x1D  # Get Instantaneous Position
SET_ACTUAL_POSITION_UNITS=0xBE

# globals for odometry - ned to keep track of the last number of counts on each wheel
LastNumber0 = 0
LastNumber1 = 0

# global reference to the serial port
ser = serial.Serial()

# routines for accessing parts of various values
def hibyte(word):
    return word/256

def lobyte(word):
    return word - (hibyte(word) * 256)

def hiword(dword):
    return dword/65536

def loword(dword):
    return dword - (hiword(dword) * 65536)

class PWDCmd:
    address = 0x0
    checksum = 0x0
    axis = 0        #Always zero
    code = 0x0
    data = [0xCC,0xCC,0xCC,0xCC,0xCC,0xCC]

class OdometricPosition:
    x = 0.
    y = 0.
    theta = 0.      # radians from 0 to 2pi

    def pr(self):
        print 'x:', self.x
        print 'y:', self.y
        print 'th (degrees):', self.theta*180.0/math.pi
        print 'th (radians):', self.theta

    def set(self,x_in,y_in,th_in):
        self.x = x_in
        self.y = y_in
        self.theta = th_in

    def reset(self):
        self.set(0.,0.,0.)

    def diffDriveKinUpdate(self, left_cm, right_cm):
        """This method takes in
               left_cm  = the number of cm the left wheel has turned
               right_cm = the number of cm the right wheel has turned
                          (forward is positive, reverse is negative)

            and it changes the calling object to reflect this motion
        """
        #
        # debugging
        #
        #print 'in Kin, l,r are', left_cm, right_cm

        # let's leave if there is nothing to do...
        if left_cm == 0 and right_cm == 0:
            return

        
        # compute the radius of curvature (ROC)
        #
        # this assumes that the robot has been traveling with constant
        # radius of curvature, which should be mostly satisfied
        # as long as we update the position every time that the velocity
        # is changed
        #
        # ROC = DELTA * (R + L) / (R - L), where DELTA is 1/2 distance between the wheels
        ROCdenominator = right_cm - left_cm
        ROCnumerator = right_cm + left_cm

        # print 'ROCdenominator is', ROCdenominator;
        # print 'ROCnumerator is', ROCnumerator;

        # if the denominator's value is too small relative to the numerator, the robot is 
        # essentially traveling in a straight line - the following ratio of 10000:1 means
        # that we will consider anything within 0.005 of a degree a straight line - 
        # likely undetectable, given all of the other sources of odometric error

        if ROCdenominator == 0 or math.fabs(ROCnumerator/float(ROCdenominator)) > 10000.0:
            # theta does not change
            self.x += math.cos(self.theta)*left_cm # left_cm ~= right_cm
            self.y += math.sin(self.theta)*left_cm

        else:  # left_cm ~= right_cm
            # otherwise, we have to figure out the ROC and the angle turned
            # the angle, theta, is computed based on the wheel that turned further
            #
            # that wheel rotated around a radius of (DELTA + fabs(ROC))
            #
            # theta = max(fabs(L),fabs(R)) / (DELTA + fabs(ROC)) 
            #
            # DELTA has to be in the same units as w1 and w0...
            ROC = DELTA * float(ROCnumerator) / float(ROCdenominator)
            #print 'ROC is', ROC
            theta = max(math.fabs(right_cm),math.fabs(left_cm)) / (DELTA + math.fabs(ROC))
            # but theta's sign is not necessarily correct...
            #
            # we will use a positive value to indicate counterclockwise rotation (left)
            # with respect to the robot's "forward-facing" pose
            #
            # the opposite (negative) rotation occurs when the left wheel has
            # traveled further
            #
            if left_cm > right_cm:
                theta = -theta

            # new conversion
            #print 'theta is', theta*180.0/math.pi;

            # first, we compute the canonical distance traveled
            # assume the robot is facing the positive x direction
            # and that positive y is to the left and the robot
            # is at (0,0)
            #
            # hooray for ASCII art!
            #
            #                ^ +x (aligned with self.theta)
            #                |
            # L=left_cm  ^   |   ^  right_cm=R
            #            |   |   |
            #     +y <---W1--+--W0--- -y (perp to self.theta)
            #                |
            #                |       DELTA = 1/2 distance from W1 to W0
            #                | -x
            #
            # The wheels W1 and W0 give rise to positive differences
            # when they are headed forward along the positive x axis
            #
            # then the center of the circle along which it's traveling
            # is at (0,ROC)
            # 
            # and the resulting position after the arc is
            # (0,ROC) + (abs(ROC)*sin(theta),-ROC*cos(theta))
            #
	    # kinematics error #2: the line below was deltax = 0.0 + math.fabs(ROC)*math.sin(theta)
	    # but it should be
	    deltax = 0.0 + ROC*math.sin(theta)
            deltay = ROC - ROC*math.cos(theta)
            #print 'deltax is', deltax
            #print 'deltay is', deltay

            # now, we need to convert this local-coordinate deltax and deltay
            # into the global frame - 
            # deltax is the change along self.theta
            # deltay is the change perpendicular to self.theta, along self.theta + pi/2
            #
            x_from_deltax = math.cos(self.theta) * deltax
            # are you joking me! the value below on the right was deltay !!
            y_from_deltax = math.sin(self.theta) * deltax
            x_from_deltay = math.cos(self.theta + math.pi/2.0) * deltay
            y_from_deltay = math.sin(self.theta + math.pi/2.0) * deltay

            # adjust position
            self.x += (x_from_deltax + x_from_deltay)
            self.y += (y_from_deltax + y_from_deltay)
            self.theta += theta

            # make sure we're less or equal to 2pi
            while self.theta > math.pi*2.:
                self.theta -= math.pi*2.
            # make sure we're greater than or equal to 0
            while self.theta < 0:
                self.theta += math.pi*2.

        #print 'current pos is'
        #self.pr();

# global value to hold the robot's odometry
gpos = OdometricPosition()

def SetCmd2(byte1, byte2):
    cmd = PWDCmd()
    #for some reason this gets changed in the class, so always reset to default
    cmd.data = [0xCC,0xCC,0xCC,0xCC,0xCC,0xCC]  
    cmd.address = byte1
    cmd.code = byte2
    cmd.checksum = ~(byte1 + byte2) + 1
    if cmd.checksum < 0:
        cmd.checksum += 256
    return cmd

def SetCmd3(byte1, byte2, data1):
    cmd = SetCmd2(byte1, byte2)
    cmd.data[0] = hibyte(data1)
    cmd.data[1] = lobyte(data1)
    cmd.checksum = ~(~(cmd.checksum-1)+ cmd.data[0] + cmd.data[1]) + 1
    if cmd.checksum < 0:
        cmd.checksum += 256
    return cmd;

def SetCmd4(byte1, byte2, data1, data2):
    cmd = SetCmd3(byte1, byte2, data1)
    cmd.data[2] = hibyte(data2)
    cmd.data[3] = lobyte(data2)
    cmd.checksum = ~(~(cmd.checksum-1)+ cmd.data[2] + cmd.data[3]) + 1
    while cmd.checksum < 0:
        cmd.checksum += 256
    while cmd.checksum > 256:
        cmd.checksum -= 256
    return cmd

def SendCmd(cmd):
    global DEBUG
    reply = ""
    if not NOSEND:
        bytes = ""
        bytes += chr(cmd.address)
        bytes += chr(cmd.checksum)
        bytes += chr(cmd.axis)
        bytes += chr(cmd.code)
        for b in cmd.data:
            if b < 0:
                b += 256
            if b != 0xCC:       # For whatever reason 0xCC is "null" -> we could fix this by keeping track of size...
                try:
                    bytes += chr(b)
                except ValueError:
                    print "ValueError: " + str(b)
                    # chr(b) Throws error now.
                    DEBUG = 1
        ser.write(bytes)
        time.sleep(0.03)        # Wait for response to be created
        reply = ser.read(size=8)
            
    if DEBUG:
        bytes = ""
        bytes += str(cmd.address)
        bytes += " "
        bytes += str(cmd.checksum)
        bytes += " "
        bytes += str(cmd.axis)
        bytes += " "
        bytes += str(cmd.code)
        bytes += " "
        for b in cmd.data:
            if b < 0:
                b += 256
            bytes += str(b) + " "
        print bytes
        if not NOSEND:
            newstring = "Reply "
            for s in reply:
                newstring += str(ord(s)) + " "
            print newstring
    return reply

def SendCmd2(byte1, byte2):
    return SendCmd(SetCmd2(byte1, byte2))

def SendCmd3(byte1, byte2, data1):
    return SendCmd(SetCmd3(byte1, byte2, data1))

def SendCmd4(byte1, byte2, data1, data2):
    return SendCmd(SetCmd4(byte1, byte2, data1, data2))

def WheelUpdate():
    SendCmd2(1,UPDATE)
    SendCmd2(0,UPDATE)

def WheelSetPower(data):
    SendCmd3(1, SET_MOTOR_COMMAND, data)
    SendCmd3(0, SET_MOTOR_COMMAND, data)

def WheelInit():
    SendCmd2(1, RESET)
    SendCmd3(1, SET_LIMIT_SWITCH_MODE, 0x0000)
    SendCmd3(1, SET_MOTOR_COMMAND, 0x0000)
    SendCmd3(1, SET_PROFILE_MODE, 0x0001)
    SendCmd4(1, SET_ACCELERATION, 0x0000, 0x00B0)
    SendCmd4(1, SET_DECELERATION, 0x0000, 0x00B0)
    SendCmd4(1, SET_JERK, 0x0329, 0x0ABB)

    SendCmd2(0, RESET)
    SendCmd3(0, SET_LIMIT_SWITCH_MODE, 0x0000)
    SendCmd3(0, SET_MOTOR_COMMAND, 0x0000)
    SendCmd3(0, SET_PROFILE_MODE, 0x0001)
    SendCmd4(0, SET_ACCELERATION, 0x0000, 0x00B0)
    SendCmd4(0, SET_DECELERATION, 0x0000, 0x00B0)
    SendCmd4(0, SET_JERK, 0x0329, 0x0ABB)

    # this seems to need to be done each time...
    # WheelSetPower(0x4CC0)

    # this seems to need to be done each time, as well...
    WheelUpdate()

def convert_counts_to_cm(counts):
    # convert to units of centimeters
    return counts * WHEEL_CIRCUMFERENCE / COUNTS_PER_REVOLUTION

def UpdateOdometry():
    # this seems mostly to return the same magnitude when the velocities are
    # identical, so that the time-delay associated with two separate
    # calls to the wheel odometry does not seem to have an effect
    # well, there is a difference of 255 some of the time... not
    # consistenly, however
    reply = SendCmd2(1, GET_COMMANDED_POSITION)
    reply2 = SendCmd2(0, GET_COMMANDED_POSITION)

    # convert number1 to the number of microsteps that have been been taken
    number1 = 0
    i = 3
    for s in reply[2:5]:
        number1 += ord(s)*(256**i)
        i -= 1
    if number1 > (0xFFFFFFFF/2): #Due to the limit of the PMD processor, we only have a ulong. 
        # So, half is pos, half is neg.
        # That is to say, we are using two's complement arithmetic
        number1 -= 0xFFFFFFFF

    # this value is reversed in order to make the differences positive
    # when the robot is moving as shown in the above ascii diagram
    number1 = -number1
    
    # convert number0 to the number of microsteps that have been taken
    number0 = 0
    i = 3
    for s in reply2[2:5]:
        number0 += ord(s)*(256**i)
        i -= 1
    if number0 > (0xFFFFFFFF/2):
        number0 -= 0xFFFFFFFF
    
    # this value is not reversed in sign to make the differences positive
    # when the robot is moving as shown in the above ascii diagram
    # it's here in case the robot is reconfigured and it needs to be changed!
    #
    # number0 = number0

    # take differences from last remembered values
    # which are the last time the odometry (pos) was updated
    global LastNumber0, LastNumber1
    diff1 = number1 - LastNumber1
    diff0 = number0 - LastNumber0

    # what if the value wraps?
    #
    # we will assume the value has wrapped if either difference diff0 or diff1
    # is greater than 3 billion... (a bit more than 2**31)
    # the register holding the counts can only measure from -2**31 to 2**31
    #
    # at 256*210 steps per revolution and a wheel diameter of 10 cm, 
    # this would cause aliasing if a wheel traveled for more than
    # 11,000 meters without an update (probably not a worry)

    # Indeed, since the robot is unlikely ever to travel 11k, I've
    # commented out this code for efficiency... I'll just 
    # keep it here for completeness

    # if math.fabs(diff1) > 3000000000:
    #     # assume it has wrapped and take the difference the other way
    #     diff1 = 2**31 - math.fabs(number1) + 2**31 - math.fabs(LastNumber1)
    #     # if the new value is negative and the old one was
    #     # positive, then LastNumber1 is > 0 but it was moving forward
    #     # so diff1 should be positive (which it is)
    #     # on the other hand, if the new value is positive and 
    #     # the old one was negative (and the diff magnitude is > 3000000000)
    #     # then it's going backwards and so should be negative
    #     if LastNumber1 < 0:
    #         diff1 = -diff1
    
    # and we include the same thing for diff0

    # if math.fabs(diff0) > 3000000000:
    #     diff0 = 2**31 - math.fabs(number0) + 2**31 - math.fabs(LastNumber0)
    #     if LastNumber0 < 0:
    #         diff0 = -diff0
    
    #print 'number0,1 are', number0, number1 
    #print 'preconverted diff0,1 are', diff0, diff1

    # save the last number of counts
    LastNumber1 = number1
    LastNumber0 = number0

    # run the kinematics

    # first, change wheel movement to centimeters
    left_wheel_cm = convert_counts_to_cm(diff1)
    right_wheel_cm = convert_counts_to_cm(diff0)
    
    # next, update the current odometer based on those motions
    gpos.diffDriveKinUpdate(left_wheel_cm, right_wheel_cm)


def SetWheelVelocities(left_cm_s, right_cm_s):
    # make sure they're within our safe range
    # if not, scale to make them so
    # I'll use 20 cm/sec as the limit...
    # for the standard base, this is a bit more than 60 degrees/sec
    # when turning in place
    #
    maxSpeed = max(left_cm_s, right_cm_s, -left_cm_s, -right_cm_s)

    SPEEDLIMIT = 20.0  # centimeters per second

    if maxSpeed > SPEEDLIMIT:
        left_cm_s *= SPEEDLIMIT/maxSpeed
        right_cm_s *= SPEEDLIMIT/maxSpeed

    # motor 1's has to be reversed in sign before sending
    left_cm_s = -left_cm_s

    #
    # these must be converted to internal units 
    # (which don't seem to be counts/second)
    #
    VELOCITY_CONVERSION = 16700.0
    left = int(left_cm_s * VELOCITY_CONVERSION)
    right = int(right_cm_s * VELOCITY_CONVERSION)

    # if we get here and somehow left or right are bigger than
    # 30 cm/s = 30*16700, we print an error message and leave
    # this could happen if the SPEEDLIMIT were changed...
    if math.fabs(left) > 30.0*VELOCITY_CONVERSION or math.fabs(right) > 30.0*VELOCITY_CONVERSION:
        print 'with left, right =', left, right
        print 'exceeding speed limits, we are not changing velocity'
        return

    # since the odometry calculations only work for 
    # time-intervals of constant ROC, we update with each
    # velocity change
    UpdateOdometry()

    # now send the command to the robot
    WheelSetPower(0x4CC0)
    SendCmd4(1, SET_VELOCITY, hiword(left), loword(left))
    SendCmd4(0, SET_VELOCITY, hiword(right), loword(right))
    WheelUpdate()
    #print 'time is', time.clock()

def SetVelocities_cm_deg(cmpersec, degpersec):
    # need to invert the kinematics to get the wheel rotations
    # in cm per second

    # the translational component is easy
    left_trans_cm_sec = cmpersec
    right_trans_cm_sec = cmpersec

    # the rotational component is not too bad 
    left_rot_cm_sec = - DELTA * (degpersec * math.pi / 180.0 )
    right_rot_cm_sec = DELTA * (degpersec * math.pi / 180.0 )

    # add them up
    left_cm_sec = left_trans_cm_sec + left_rot_cm_sec
    right_cm_sec = right_trans_cm_sec + right_rot_cm_sec

    # send off to robot
    SetWheelVelocities(left_cm_sec, right_cm_sec)

def Go(cmpersec, degpersec=0.0):
    SetVelocities_cm_deg(cmpersec,degpersec) 

def OpenSerialPort(PORT):
    if ser.isOpen():
        print "ERROR: serial port already open"
        sys.exit(-1)

    # this should be improved...
    if sys.platform != 'win32':
        ser.port = PORT
        ser.baudrate = 230400
    else:
        PORT = int(PORT)
        ser.port = PORT-1
        ser.baudrate = 250000
    ser.timeout = 0
    ser.open()
    if not ser.isOpen():
        print "ERROR: serial port did not open"
        sys.exit(-1)

def InitializeRobot(PORT):
    OpenSerialPort(PORT)
    WheelInit()
    UpdateOdometry()
    global gpos
    gpos.x=gpos.y=gpos.theta=0
    # sometimes they're not 0 at this point for reasons unknown...

def SetOdometry(x,y,thr):
    
    global gpos
    gpos.x = x
    gpos.y = y
    gpos.theta = thr

def GetOdometry():
    UpdateOdometry()
    # theta is in radians...
    return (gpos.x,gpos.y,gpos.theta)

# "Where" for the lazy...
def Wh():
    return GetOdometry()
    
def PowerOff():
    # this will update the odometry one last time...
    # it will also briefly turn the power on (if it wasn't already)
    SetWheelVelocities(0.0,0.0)
    # this will turn the power off  
    WheelSetPower(0x0)
    WheelUpdate()

def CloseRobot():
    PowerOff()
    ser.close()


