# er2real python interface
#
# The er2 class here (should have/has) the
#   same interface as the er2 class that controls the
#   real robot...
#
# r = er2real.ER2(17)  (or whatever the port is...)
#                      you can also specify a map file...
#
# Go(cmpersec, degpersec=0.0) - sets translational, rotational velocities
#
# Wh() - returns a tuple of (x,y,th) odometry
#
# PowerOff() - turns off power to the motors Go(...) turns it back on
#              note there is a difference between Go(0) and PowerOff():
#              Go(0) will actively hold a velocity of 0 (e.g., on hills)
#              PowerOff() lets the wheels float (e.g., roll down a hill)
#
# CloseRobot() - powers off and closes the serial port


import time
import math
import random
import sys
import serial

# used in two places... perhaps this should be fixed?
GLOBAL_DELTA = 19.0   # 1/2 the distance between the wheels in CM

# stuff taken from the C++ interface
UPDATE                  = 0x1A
SET_MOTOR_COMMAND       = 0x77
RESET                   = 0x39
SET_LIMIT_SWITCH_MODE   = 0x80
SET_PROFILE_MODE        = 0xA0
SET_ACCELERATION        = 0x90
SET_DECELERATION        = 0x91
SET_JERK                = 0x13
SET_VELOCITY            = 0x11
GET_COMMANDED_POSITION  = 0x1D  # Get Instantaneous Position
SET_ACTUAL_POSITION_UNITS=0xBE


# routines for accessing parts of various values
def hibyte(word):
    return word/256

def lobyte(word):
    return word - (hibyte(word) * 256)

def hiword(dword):
    return dword/65536

def loword(dword):
    return dword - (hiword(dword) * 65536)

class PWDCmd:
    address = 0x0
    checksum = 0x0
    axis = 0        #Always zero
    code = 0x0
    data = [0xCC,0xCC,0xCC,0xCC,0xCC,0xCC]


class ER2OdometricPosition:

    def __init__(self, x_in=0., y_in=0., thr_in=0.):
        self.x = x_in
        self.y = y_in
        self.thr = thr_in # radians from 0 to 2pi - change to thr sometime...
        self.DELTA = GLOBAL_DELTA

    def pr(self):
        print 'x:', self.x
        print 'y:', self.y
        print 'th (degrees):', self.thr*180.0/math.pi
        print 'th (radians):', self.thr

    def set(self,x_in,y_in,thr_in):
        self.x = x_in
        self.y = y_in
        self.thr = thr_in

    def reset(self):
        self.set(0.,0.,0.)

    def diffDriveKinUpdate(self, left_cm, right_cm):
        """This method takes in
               left_cm  = the number of cm the left wheel has turned
               right_cm = the number of cm the right wheel has turned
                          (forward is positive, reverse is negative)

            and it changes the calling object to reflect this motion
        """
        #
        # debugging
        #
        #print 'in Kin, l,r are', left_cm, right_cm

        # let's leave if there is nothing to do...
        if left_cm == 0 and right_cm == 0:
            return

        
        # compute the radius of curvature (ROC)
        #
        # this assumes that the robot has been traveling with constant
        # radius of curvature, which should be mostly satisfied
        # as long as we update the position every time that the velocity
        # is changed
        #
        # ROC = DELTA * (R + L) / (R - L), where DELTA is 1/2 distance between the wheels
        ROCdenominator = right_cm - left_cm
        ROCnumerator = right_cm + left_cm

        # if the denominator's value is too small relative to the numerator, the robot is 
        # essentially traveling in a straight line - the following ratio of 10000:1 means
        # that we will consider anything within 0.005 of a degree a straight line - 
        # likely undetectable, given all of the other sources of odometric error

        if ROCdenominator == 0 or math.fabs(ROCnumerator/float(ROCdenominator)) > 10000.0:
            # theta does not change
            self.x += math.cos(self.thr)*left_cm # left_cm ~= right_cm
            self.y += math.sin(self.thr)*left_cm

        else:  # left_cm ~= right_cm
            #global gvel_left_cmpersec, gvel_right_cmpersec, gtime
            #print 'gvel:l,r,time are', gvel_left_cmpersec, gvel_right_cmpersec, gtime
	    #print 'gpos before is', self.x, self.y, self.theta*180.0/math.pi
	    #print 'right_cm, left_cm are', right_cm, left_cm
            #print 'ROCden, ROCnum are', ROCdenominator, ROCnumerator
	    #
            # otherwise, we have to figure out the ROC and the angle turned
            # the angle, theta, is computed based on the wheel that turned further
            #
            # that wheel rotated around a radius of (DELTA + fabs(ROC))
            #
            # theta = max(fabs(L),fabs(R)) / (DELTA + fabs(ROC)) 
            #
            # DELTA has to be in the same units as w1 and w0...
            ROC = self.DELTA * float(ROCnumerator) / float(ROCdenominator)
            #print 'ROC is', ROC
            d_thr = max(math.fabs(right_cm),math.fabs(left_cm)) / (self.DELTA + math.fabs(ROC))
            # but theta's sign is not necessarily correct...
            #
            # we will use a positive value to indicate counterclockwise rotation (left)
            # with respect to the robot's "forward-facing" pose
            #
            # the opposite (negative) rotation occurs when the left wheel has
            # traveled further
            #
            if left_cm > right_cm:
                d_thr = -d_thr

            # new conversion
            #print 'theta is', theta*180.0/math.pi;

            # first, we compute the canonical distance traveled
            # assume the robot is facing the positive x direction
            # and that positive y is to the left and the robot
            # is at (0,0)
            #
            # hooray for ASCII art!
            #
            #                ^ +x (aligned with self.theta)
            #                |
            # L=left_cm  ^   |   ^  right_cm=R
            #            |   |   |
            #     +y <---W1--+--W0--- -y (perp to self.theta)
            #                |
            #                |       DELTA = 1/2 distance from W1 to W0
            #                | -x
            #
            # The wheels W1 and W0 give rise to positive differences
            # when they are headed forward along the positive x axis
            #
            # then the center of the circle along which it's traveling
            # is at (0,ROC)
            # 
            # and the resulting position after the arc is
            # (0,ROC) + (ROC*sin(theta),-ROC*cos(theta))
            #
	    # kinematics error #2: used absolute value below...
            deltax = 0.0 + ROC*math.sin(d_thr)
            deltay = ROC - ROC*math.cos(d_thr)
            #print 'local deltax and deltay are', deltax, deltay

            # now, we need to convert this local-coordinate deltax and deltay
            # into the global frame - 
            # deltax is the change along self.theta
            # deltay is the change perpendicular to self.theta, along self.theta + pi/2
            #
            x_from_deltax = math.cos(self.thr) * deltax
            # are you joking me!! the value on the right below was deltay - ouch!
            y_from_deltax = math.sin(self.thr) * deltax
            x_from_deltay = math.cos(self.thr + math.pi/2.0) * deltay
            y_from_deltay = math.sin(self.thr + math.pi/2.0) * deltay

            #print 'global deltax and deltay are', (x_from_deltax + x_from_deltay), (y_from_deltax + y_from_deltay)

            # adjust position
            self.x += (x_from_deltax + x_from_deltay)
            self.y += (y_from_deltax + y_from_deltay)
            self.thr += d_thr

            # make sure we're less or equal to 2pi
            while self.thr > math.pi*2.:
                self.thr -= math.pi*2.
            # make sure we're greater than or equal to 0
            while self.thr < 0:
                self.thr += math.pi*2.

            #print for debugging...
            #print 'l,r,ROC,thd are', left_cm, right_cm, ROC, self.theta*180.0/3.14159
	    #print 'gpos after is', self.x, self.y, self.theta*180.0/math.pi, '\n'

        #print 'current pos is'
        #self.pr();


class ER2SensorSnapshot:

    def __init__(self, sensorData=None):
        if sensorData == None:
            self.sensorData = []
        else:
            self.sensorData = sensorData


class ER2:

    def serialStatus(self):
        print 'serial info is', self.ser

    def __init__(self, PORT, mapfile=None):
        self.ser = None # no serial port yet!
        self.NOSEND = 0
        self.DEBUG = 0
        # Kinematic constants
        #WHEEL_CIRCUMFERENCE = math.pi * 10    
        # Evolution's wheels are 10 cm in diameter
        #COUNTS_PER_REVOLUTION = 256*210.0     
        # Empirically determined for the standard ER1
        self.LastNumber0 = 0
        self.LastNumber1 = 0  # last value of counts for each wheel
        self.WHEEL_CIRCUMFERENCE = math.pi * 10    # Evolution's wheels are 10 cm in diameter
        self.COUNTS_PER_REVOLUTION = 256*210.0     # Empirically determined for the standard ER1
        self.DELTA = GLOBAL_DELTA   # cm from wheel to center of wheels: 19.0
        # DELTA is needed to convert from deg per sec to wheel velocities...
        
        # global velocities for the robot
        self.gvel_left_cmpersec = 0.0
        self.gvel_right_cmpersec = 0.0
        
        # global value to hold the robot's odometric position
        self.gpos = ER2OdometricPosition()
        # pose and sensor history with timestamps
        # WORRY ABOUT HISTORY LATER
	# self.fullHistory = []
        # each entry is (x,y,thr,(sensor tuple),timestamp)
        #self.fullHistory.append( (0.0,0.0,0.0,(),0.0,(0.0,0.0,0.0)) )
        # update will update this full history...

        # global timestamp: last position-updated time
        # we assume the velocity is constant between updates of this variable
        # this time.clock() needs to be called to start the clock...
        # WORRY ABOUT TIME LATER -- do we need it?
        #self.lastUpdatedTime = time.clock() # initialize to about zero
        # get the map - must be done before the call to self.InitializeRobot,
        # which calls update, which needs the map...
        self.map = None
        # initialize robot with port... Windows adds one...
        # it will get subtracted in OpenSerialPort
        self.PORT = PORT
        self.InitializeRobot(self.PORT)

    def motionErrorModelVelocities(self):
	# returns a pair of values: the "actual" right and left
	# wheel velocities in cm/sec

	# returning self.gvel_left_cmpersec and self.gvel_right_cmpersec
	# means there is no error at all

	return  (self.gvel_left_cmpersec  + random.gauss(0.0,1.0), \
                 self.gvel_right_cmpersec + random.gauss(0.0,1.0))

	# a real lean!
	#return  (self.gvel_left_cmpersec + 1.0, self.gvel_right_cmpersec - 1.0)


    def updateNOTUSED(self, deltaTimeSeconds=None, updateOnlyPose=False):
        # we need to move everything forward in time
        # we will assume now that this is being called fast enough so
        # that all relevent events (crashes, etc.) are being caught
        # THIS MAY ALWAYS HAVE TO BE ASSUMED... !!

        # it does seem a good idea to be able to pass in an 
        # artificial deltaTimeSeconds, however. 
        # this will allow for realistic, non-real-time simulations
        # if lots of computation is required...

        # what's the "right" way to handle artificial and real time
        # together? (probably just not to do it)

        # we'll worry about turning sensors on/off later...

        # OK, we know the time delta - now, how about the real vs.
	# the noisy robot location?

	# Well, let's assume that the estimated, noisy robot position
	# is the location obtained by integrating the commanded
	# velocities over time  (this, indeed, will be noisy)

        # update the [potential!] pose changes
        left_wheel_cm = self.gvel_left_cmpersec * deltaTimeSeconds
        right_wheel_cm = self.gvel_right_cmpersec * deltaTimeSeconds
        
        # next, update the current odometer based on those motions
        self.gpos.diffDriveKinUpdate(left_wheel_cm, right_wheel_cm)

	# Now, what about the REAL robot position - that is going 
	# to be some other location, based on nearby wheel velocities
	# we can imagine that slippage etc. can be modeled as
	# simply a (perhaps small) change in the commanded velocity
	# to each wheel. For the moment, we'll add a Gaussian-sampled
	# delta to each wheel independently, with the distribution
	# width proportional to wheel velocity
	# This way 0 velocity is really 0 velocity, which is fair.
        actual_left_cmpersec, actual_right_cmpersec = self.motionErrorModelVelocities()
        left_wheel_cmReal = actual_left_cmpersec * deltaTimeSeconds
        right_wheel_cmReal = actual_right_cmpersec * deltaTimeSeconds
        
        # next, update the current odometer based on those motions
        self.gposReal.diffDriveKinUpdate(left_wheel_cmReal, right_wheel_cmReal)

        # next, check for collisions...
        C = self.map.findER1collisions(self.gposReal.x, \
			self.gposReal.y, self.gposReal.thr)

        # if collision, we assume it did not move from the last position
        if C:
            lastState = self.fullHistory[-1]
            x = lastState[0] # should we reset these?
            y = lastState[1] # the odometric (not "real") info?
            thr = lastState[2]
	    realPos = lastState[5]
	    #self.gpos.set(x,y,thr)
            self.gposReal.set(realPos[0], realPos[1], realPos[2])
            print 'Collision!'
            stalled = True
        else:
            stalled = False

        # next, update active sensors
        # no notion of "active" quite yet...
        # note that this uses the updated self.gpos, which is
        # good, because we want to return the sensor values
        # of the previous pose if, indeed, the robot is stuck
        if updateOnlyPose == False:
            R = self.map.findER1ranges(self.gposReal.x, \
			    self.gposReal.y, self.gposReal.thr)
        else:
            R = []

        # put together pose and sensor data with timestamp
        # add to history
        x = self.gpos.x
        y = self.gpos.y
        thr = self.gpos.thr
        sensorTuple = (R,stalled)
	realPos = (self.gposReal.x, self.gposReal.y, self.gposReal.thr)
        # newtime is the timestamp
        self.fullHistory.append( (x,y,thr,sensorTuple,newtime,realPos) )

        # regardless, we set the lastUpdatedTime
        self.lastUpdatedTime = newtime


    def clearHistory():
        if len(self.poseHistory) > 0:
            self.poseHistory = self.poseHistory[-1:]
        # otherwise, it's already empty
    
    def SetVelocities_cm_deg(self, cmpersec, degpersec):
        # need to invert the kinematics to get the wheel rotations
        # in cm per second
    
        # the translational component is easy
        left_trans_cm_sec = cmpersec
        right_trans_cm_sec = cmpersec
    
        # the rotational component is not too bad 
        # note that if the translational component is 0, this
        # allows for a maximum of about 60 degrees per second...
    
        # ok, I think this does not combine rotation and translation
        # in the most natural way
        DELTA = self.DELTA
        left_rot_cm_sec = - self.DELTA * (degpersec * math.pi / 180.0 )
        right_rot_cm_sec = self.DELTA * (degpersec * math.pi / 180.0 )
    
        print 'cm, deg', cmpersec, ',', degpersec
        print 'left,right trans', left_trans_cm_sec, ',', right_trans_cm_sec
    
        # instead, we should add a constant that makes the smaller of
        # these equal to zero - that way the slower wheel will travel
        # the desired translational amount, and the rotation will
        # be taken up by the faster wheel...
    
        # since they're of opposite sign, this will work
        #adjust = max(left_rot_cm_sec, right_rot_cm_sec)
        #left_rot_cm_sec += adjust
        #right_rot_cm_sec += adjust
        print 'left,right rot', left_rot_cm_sec, ',', right_rot_cm_sec
    
        # add them up
        left_cm_sec = left_trans_cm_sec + left_rot_cm_sec
        right_cm_sec = right_trans_cm_sec + right_rot_cm_sec
        print 'left,right total', left_cm_sec, ',', right_cm_sec
    
        # send off to robot
        self.SetWheelVelocities(left_cm_sec, right_cm_sec)


    def SetWheelVelocities(self, left_cm_sec, right_cm_sec):
        # we need to update the odometry because whenever the velocity changes,
        # so does the interpretation of the distance traveled by each wheel...
        #self.update(updateOnlyPose=True)
        # note that this call to UpdateOdometry has to be done FIRST,
        # so that the last time step is interpreted correctly

        #global gpos, gvel_right_cmpersec, gvel_left_cmpersec
        # why change if there's no change to make?
        if self.gvel_right_cmpersec == right_cm_sec and self.gvel_left_cmpersec == left_cm_sec: return

        # we can comment out speed limits, if we want...
        # but they're here for now!
        maxSpeed = max(left_cm_sec, right_cm_sec, -left_cm_sec, -right_cm_sec)
    
        # much bigger speed limit here in simulation... (vs. 20 or so)
        SPEEDLIMIT = 20.0  # centimeters per second
    
        if maxSpeed > SPEEDLIMIT:
            left_cm_sec *= SPEEDLIMIT/maxSpeed
            right_cm_sec *= SPEEDLIMIT/maxSpeed
    
        self.gvel_right_cmpersec = right_cm_sec
        self.gvel_left_cmpersec = left_cm_sec

        # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        #
        # HERE IS WHERE MOTOR DIRECTION IS DETERMINED
        # motor 1's has to be reversed in sign before sending
        # also, we need to change to a slightly different name
        #
        # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        left_cm_s = +left_cm_sec
        right_cm_s = -right_cm_sec
    
        #
        # these must be converted to internal units 
        # (which don't seem to be counts/second)
        #
        VELOCITY_CONVERSION = 16700.0
        left = int(left_cm_s * VELOCITY_CONVERSION)
        right = int(right_cm_s * VELOCITY_CONVERSION)
    
        # if we get here and somehow left or right are bigger than
        # 30 cm/s = 30*16700, we print an error message and leave
        # this could happen if the SPEEDLIMIT were changed...
        if math.fabs(left) > 30.0*VELOCITY_CONVERSION or math.fabs(right) > 30.0*VELOCITY_CONVERSION:
            print 'with left, right =', left, right
            print 'exceeding speed limits, we are not changing velocity'
            return
    
        # since the odometry calculations only work for 
        # time-intervals of constant ROC, we update with each
        # velocity change
        self.UpdateOdometry()
    
        # now send the command to the robot
        self.WheelSetPower(0x4CC0)
        self.SendCmd4(1, SET_VELOCITY, hiword(left), loword(left))
        self.SendCmd4(0, SET_VELOCITY, hiword(right), loword(right))
        self.WheelUpdate()
        #print 'time is', time.clock()

    def Wh(self):
        self.UpdateOdometry()
        return (self.gpos.x, self.gpos.y, self.gpos.thr)
    
    def Go(self, cmpersec, degpersec=0.0):
        self.SetVelocities_cm_deg(cmpersec,degpersec) 

    def OpenSerialPort(self, PORT):
        if self.ser == None:
            self.ser = serial.Serial()

        if self.ser.isOpen():
            print "ERROR: serial port already open"
            sys.exit(-1)
    
        # this should be improved...
        if sys.platform != 'win32':
            self.ser.port = PORT
            self.ser.baudrate = 230400
        else:  # Windows!
            PORT = int(PORT)
            self.ser.port = PORT-1
            self.ser.baudrate = 250000
        self.ser.timeout = 0
        self.ser.open()
        if not self.ser.isOpen():
            print "ERROR: serial port did not open"
            sys.exit(-1)
    
    
    def SetCmd2(self, byte1, byte2):
        cmd = PWDCmd()
        #for some reason this gets changed in the class, so always reset to default
        cmd.data = [0xCC,0xCC,0xCC,0xCC,0xCC,0xCC]  
        cmd.address = byte1
        cmd.code = byte2
        cmd.checksum = ~(byte1 + byte2) + 1
        if cmd.checksum < 0:
            cmd.checksum += 256
        return cmd
    
    def SetCmd3(self, byte1, byte2, data1):
        cmd = self.SetCmd2(byte1, byte2)
        cmd.data[0] = hibyte(data1)
        cmd.data[1] = lobyte(data1)
        cmd.checksum = ~(~(cmd.checksum-1)+ cmd.data[0] + cmd.data[1]) + 1
        if cmd.checksum < 0:
            cmd.checksum += 256
        return cmd;
    
    def SetCmd4(self, byte1, byte2, data1, data2):
        cmd = self.SetCmd3(byte1, byte2, data1)
        cmd.data[2] = hibyte(data2)
        cmd.data[3] = lobyte(data2)
        cmd.checksum = ~(~(cmd.checksum-1)+ cmd.data[2] + cmd.data[3]) + 1
        while cmd.checksum < 0:
            cmd.checksum += 256
        while cmd.checksum > 256:
            cmd.checksum -= 256
        return cmd
    
    def SendCmd(self, cmd):
        reply = ""
        if not self.NOSEND:
            bytes = ""
            bytes += chr(cmd.address)
            bytes += chr(cmd.checksum)
            bytes += chr(cmd.axis)
            bytes += chr(cmd.code)
            for b in cmd.data:
                if b < 0:
                    b += 256
                if b != 0xCC:       # For whatever reason 0xCC is "null" -> we could fix this by keeping track of size...
                    try:
                        bytes += chr(b)
                    except ValueError:
                        print "ValueError: " + str(b)
                        # chr(b) Throws error now.
                        self.DEBUG = 1
            self.ser.write(bytes)
            time.sleep(0.03)        # Wait for response to be created
            reply = self.ser.read(size=8)
                
        if self.DEBUG:
            bytes = ""
            bytes += str(cmd.address)
            bytes += " "
            bytes += str(cmd.checksum)
            bytes += " "
            bytes += str(cmd.axis)
            bytes += " "
            bytes += str(cmd.code)
            bytes += " "
            for b in cmd.data:
                if b < 0:
                    b += 256
                bytes += str(b) + " "
            print bytes
            if not self.NOSEND:
                newstring = "Reply "
                for s in reply:
                    newstring += str(ord(s)) + " "
                print newstring
        return reply
    
    def SendCmd2(self, byte1, byte2):
        return self.SendCmd(self.SetCmd2(byte1, byte2))
    
    def SendCmd3(self, byte1, byte2, data1):
        return self.SendCmd(self.SetCmd3(byte1, byte2, data1))
    
    def SendCmd4(self, byte1, byte2, data1, data2):
        return self.SendCmd(self.SetCmd4(byte1, byte2, data1, data2))
    
    def WheelUpdate(self):
        self.SendCmd2(1,UPDATE)
        self.SendCmd2(0,UPDATE)
    
    def WheelSetPower(self, data):
        self.SendCmd3(1, SET_MOTOR_COMMAND, data)
        self.SendCmd3(0, SET_MOTOR_COMMAND, data)
    
    def WheelInit(self):
        self.SendCmd2(1, RESET)
        self.SendCmd3(1, SET_LIMIT_SWITCH_MODE, 0x0000)
        self.SendCmd3(1, SET_MOTOR_COMMAND, 0x0000)
        self.SendCmd3(1, SET_PROFILE_MODE, 0x0001)
        self.SendCmd4(1, SET_ACCELERATION, 0x0000, 0x00B0)
        self.SendCmd4(1, SET_DECELERATION, 0x0000, 0x00B0)
        self.SendCmd4(1, SET_JERK, 0x0329, 0x0ABB)
    
        self.SendCmd2(0, RESET)
        self.SendCmd3(0, SET_LIMIT_SWITCH_MODE, 0x0000)
        self.SendCmd3(0, SET_MOTOR_COMMAND, 0x0000)
        self.SendCmd3(0, SET_PROFILE_MODE, 0x0001)
        self.SendCmd4(0, SET_ACCELERATION, 0x0000, 0x00B0)
        self.SendCmd4(0, SET_DECELERATION, 0x0000, 0x00B0)
        self.SendCmd4(0, SET_JERK, 0x0329, 0x0ABB)
    
        # this seems to need to be done each time...
        # WheelSetPower(0x4CC0)
    
        # this seems to need to be done each time, as well...
        self.WheelUpdate()
    
    def convert_counts_to_cm(self, counts):
        # convert to units of centimeters
        return counts * self.WHEEL_CIRCUMFERENCE / self.COUNTS_PER_REVOLUTION
    
    def UpdateOdometry(self):
        # this seems mostly to return the same magnitude when the velocities are
        # identical, so that the time-delay associated with two separate
        # calls to the wheel odometry does not seem to have an effect
        # well, there is a difference of 255 some of the time... not
        # consistenly, however
        reply = self.SendCmd2(1, GET_COMMANDED_POSITION)
        reply2 = self.SendCmd2(0, GET_COMMANDED_POSITION)
    
        # convert number1 to the number of microsteps that have been been taken
        number1 = 0
        i = 3
        for s in reply[2:5]:
            number1 += ord(s)*(256**i)
            i -= 1
        if number1 > (0xFFFFFFFF/2): #Due to the limit of the PMD processor, we only have a ulong. 
            # So, half is pos, half is neg.
            # That is to say, we are using two's complement arithmetic
            number1 -= 0xFFFFFFFF
    
        # this value is reversed in order to make the differences positive
        # when the robot is moving as shown in the above ascii diagram
        number1 = -number1
        
        # convert number0 to the number of microsteps that have been taken
        number0 = 0
        i = 3
        for s in reply2[2:5]:
            number0 += ord(s)*(256**i)
            i -= 1
        if number0 > (0xFFFFFFFF/2):
            number0 -= 0xFFFFFFFF
        
        # this value is not reversed in sign to make the differences positive
        # when the robot is moving as shown in the above ascii diagram
        # it's here in case the robot is reconfigured and it needs to be changed!
        #
        # number0 = number0
    
        # take differences from last remembered values
        # which are the last time the odometry (pos) was updated
        diff1 = number1 - self.LastNumber1
        diff0 = number0 - self.LastNumber0
    
        # what if the value wraps?
        #
        # we will assume the value has wrapped if either difference diff0 or diff1
        # is greater than 3 billion... (a bit more than 2**31)
        # the register holding the counts can only measure from -2**31 to 2**31
        #
        # at 256*210 steps per revolution and a wheel diameter of 10 cm, 
        # this would cause aliasing if a wheel traveled for more than
        # 11,000 meters without an update (probably not a worry)
    
        # Indeed, since the robot is unlikely ever to travel 11k, I've
        # commented out this code for efficiency... I'll just 
        # keep it here for completeness
    
        # if math.fabs(diff1) > 3000000000:
        #     # assume it has wrapped and take the difference the other way
        #     diff1 = 2**31 - math.fabs(number1) + 2**31 - math.fabs(LastNumber1)
        #     # if the new value is negative and the old one was
        #     # positive, then LastNumber1 is > 0 but it was moving forward
        #     # so diff1 should be positive (which it is)
        #     # on the other hand, if the new value is positive and 
        #     # the old one was negative (and the diff magnitude is > 3000000000)
        #     # then it's going backwards and so should be negative
        #     if LastNumber1 < 0:
        #         diff1 = -diff1
        
        # and we include the same thing for diff0
    
        # if math.fabs(diff0) > 3000000000:
        #     diff0 = 2**31 - math.fabs(number0) + 2**31 - math.fabs(LastNumber0)
        #     if LastNumber0 < 0:
        #         diff0 = -diff0
        
        #print 'number0,1 are', number0, number1 
        #print 'preconverted diff0,1 are', diff0, diff1
    
        # save the last number of counts
        self.LastNumber1 = number1
        self.LastNumber0 = number0
    
        # run the kinematics
    
        # first, change wheel movement to centimeters
        left_wheel_cm = self.convert_counts_to_cm(diff1)
        right_wheel_cm = self.convert_counts_to_cm(diff0)
        
        # next, update the current odometer based on those motions
        self.gpos.diffDriveKinUpdate(left_wheel_cm, right_wheel_cm)
    
    

    # keep the optional port number just for compatibility - it gets ignored...
    def InitializeRobot(self, PORT):
        self.OpenSerialPort(PORT)
        self.WheelInit()
        self.UpdateOdometry()
        self.gpos.reset()
        # sometimes they're not 0 at this point for reasons unknown...
        self.gvel_right_cmpersec = 0.0
        self.gvel_left_cmpersec = 0.0

    
    def SetOdometry(self, x,y,thr):
        #global gpos, gtime
        self.gpos.set(x,y,thr)
        # reset global time (gtime) via call time.clock() so that the next update
        # is from this point in time and not some past point in time...
        self.gtime = time.clock()
        # note that in the real robot, this would have to
        # do a bit more in order to get the relative motor counts 
        # updated to the current time, as well...
    
    def GetOdometry(self):
        self.UpdateOdometry()
        return (self.gpos.x, self.gpos.y, self.gpos.thr)

    def GetAllDataFrom( self, xp, yp, thrp ):
        # returns data for item (particle) at the input pose
        # first, check for in bounds!
        numhits = self.map.numHits(xp,yp,thrp)
        R = self.map.findER1ranges(xp, yp, thrp)
        return (R,numhits)

    def GetAllData(self):
        # update sensors, as well!
        self.update(updateOnlyPose=False)
        # return the last tuple...
        return self.fullHistory[-1]
    
    # "Where" for the lazy...
    def Wh(self):
        return self.GetOdometry()
        
    def PowerOff(self):
        # this will update the odometry one last time...
        # it will also briefly turn the power on (if it wasn't already)
        self.SetWheelVelocities(0.0,0.0)
        self.WheelSetPower(0x0)
        self.WheelUpdate()
    
    
    def CloseRobot(self):
        # want to keep APIs the same for sim and real robots
        self.PowerOff()
        if self.ser != None and self.ser.isOpen():
            self.ser.close()
        self.ser = None


