# er2sim python interface
#
# The er2 class here (should have/has) the
#   same interface as the er2 class that controls the
#   real robot...
#
# InitializeRobot(PORT) - (not really needed at all)
#
# Go(cmpersec, degpersec=0.0) - sets translational, rotational velocities
#
# Wh() - returns a tuple of (x,y,th) odometry
#
# PowerOff() - turns off power to the motors Go(...) turns it back on
#              note there is a difference between Go(0) and PowerOff():
#              Go(0) will actively hold a velocity of 0 (e.g., on hills)
#              PowerOff() lets the wheels float (e.g., roll down a hill)
#
# CloseRobot() - powers off and closes the serial port


import time
import math
import random
import sys
import sensorSimulator as sim

GLOBAL_DELTA = 19.0   # 1/2 the distance between the wheels

class ER2OdometricPosition:

    def __init__(self, x_in=0., y_in=0., thr_in=0.):
        self.x = x_in
        self.y = y_in
        self.thr = thr_in # radians from 0 to 2pi - change to thr sometime...
        self.DELTA = GLOBAL_DELTA

    def pr(self):
        print 'x:', self.x
        print 'y:', self.y
        print 'th (degrees):', self.thr*180.0/math.pi
        print 'th (radians):', self.thr

    def set(self,x_in,y_in,thr_in):
        self.x = x_in
        self.y = y_in
        self.thr = thr_in

    def reset(self):
        self.set(0.,0.,0.)

    def diffDriveKinUpdate(self, left_cm, right_cm):
        """This method takes in
               left_cm  = the number of cm the left wheel has turned
               right_cm = the number of cm the right wheel has turned
                          (forward is positive, reverse is negative)

            and it changes the calling object to reflect this motion
        """
        #
        # debugging
        #
        #print 'in Kin, l,r are', left_cm, right_cm

        # let's leave if there is nothing to do...
        if left_cm == 0 and right_cm == 0:
            return

        
        # compute the radius of curvature (ROC)
        #
        # this assumes that the robot has been traveling with constant
        # radius of curvature, which should be mostly satisfied
        # as long as we update the position every time that the velocity
        # is changed
        #
        # ROC = DELTA * (R + L) / (R - L), where DELTA is 1/2 distance between the wheels
        ROCdenominator = right_cm - left_cm
        ROCnumerator = right_cm + left_cm

        # if the denominator's value is too small relative to the numerator, the robot is 
        # essentially traveling in a straight line - the following ratio of 10000:1 means
        # that we will consider anything within 0.005 of a degree a straight line - 
        # likely undetectable, given all of the other sources of odometric error

        if ROCdenominator == 0 or math.fabs(ROCnumerator/float(ROCdenominator)) > 10000.0:
            # theta does not change
            self.x += math.cos(self.thr)*left_cm # left_cm ~= right_cm
            self.y += math.sin(self.thr)*left_cm

        else:  # left_cm ~= right_cm
            #global gvel_left_cmpersec, gvel_right_cmpersec, gtime
            #print 'gvel:l,r,time are', gvel_left_cmpersec, gvel_right_cmpersec, gtime
	    #print 'gpos before is', self.x, self.y, self.theta*180.0/math.pi
	    #print 'right_cm, left_cm are', right_cm, left_cm
            #print 'ROCden, ROCnum are', ROCdenominator, ROCnumerator
	    #
            # otherwise, we have to figure out the ROC and the angle turned
            # the angle, theta, is computed based on the wheel that turned further
            #
            # that wheel rotated around a radius of (DELTA + fabs(ROC))
            #
            # theta = max(fabs(L),fabs(R)) / (DELTA + fabs(ROC)) 
            #
            # DELTA has to be in the same units as w1 and w0...
            ROC = self.DELTA * float(ROCnumerator) / float(ROCdenominator)
            #print 'ROC is', ROC
            d_thr = max(math.fabs(right_cm),math.fabs(left_cm)) / (self.DELTA + math.fabs(ROC))
            # but theta's sign is not necessarily correct...
            #
            # we will use a positive value to indicate counterclockwise rotation (left)
            # with respect to the robot's "forward-facing" pose
            #
            # the opposite (negative) rotation occurs when the left wheel has
            # traveled further
            #
            if left_cm > right_cm:
                d_thr = -d_thr

            # new conversion
            #print 'theta is', theta*180.0/math.pi;

            # first, we compute the canonical distance traveled
            # assume the robot is facing the positive x direction
            # and that positive y is to the left and the robot
            # is at (0,0)
            #
            # hooray for ASCII art!
            #
            #                ^ +x (aligned with self.theta)
            #                |
            # L=left_cm  ^   |   ^  right_cm=R
            #            |   |   |
            #     +y <---W1--+--W0--- -y (perp to self.theta)
            #                |
            #                |       DELTA = 1/2 distance from W1 to W0
            #                | -x
            #
            # The wheels W1 and W0 give rise to positive differences
            # when they are headed forward along the positive x axis
            #
            # then the center of the circle along which it's traveling
            # is at (0,ROC)
            # 
            # and the resulting position after the arc is
            # (0,ROC) + (ROC*sin(theta),-ROC*cos(theta))
            #
	    # kinematics error #2: used absolute value below...
            deltax = 0.0 + ROC*math.sin(d_thr)
            deltay = ROC - ROC*math.cos(d_thr)
            #print 'local deltax and deltay are', deltax, deltay

            # now, we need to convert this local-coordinate deltax and deltay
            # into the global frame - 
            # deltax is the change along self.theta
            # deltay is the change perpendicular to self.theta, along self.theta + pi/2
            #
            x_from_deltax = math.cos(self.thr) * deltax
            # are you joking me!! the value on the right below was deltay - ouch!
            y_from_deltax = math.sin(self.thr) * deltax
            x_from_deltay = math.cos(self.thr + math.pi/2.0) * deltay
            y_from_deltay = math.sin(self.thr + math.pi/2.0) * deltay

            #print 'global deltax and deltay are', (x_from_deltax + x_from_deltay), (y_from_deltax + y_from_deltay)

            # adjust position
            self.x += (x_from_deltax + x_from_deltay)
            self.y += (y_from_deltax + y_from_deltay)
            self.thr += d_thr

            # make sure we're less or equal to 2pi
            while self.thr > math.pi*2.:
                self.thr -= math.pi*2.
            # make sure we're greater than or equal to 0
            while self.thr < 0:
                self.thr += math.pi*2.

            #print for debugging...
            #print 'l,r,ROC,thd are', left_cm, right_cm, ROC, self.theta*180.0/3.14159
	    #print 'gpos after is', self.x, self.y, self.theta*180.0/math.pi, '\n'

        #print 'current pos is'
        #self.pr();



class ER2:

    def __init__(self, PORT, mapfile=None):
        # Kinematic constants
        #WHEEL_CIRCUMFERENCE = math.pi * 10    
        # Evolution's wheels are 10 cm in diameter
        #COUNTS_PER_REVOLUTION = 256*210.0     
        # Empirically determined for the standard ER1
        global GLOBAL_DELTA  # 19.0
        self.DELTA = GLOBAL_DELTA  # 1/2 the distance bewteen the wheels = ~19cm
        # DELTA is needed to convert from deg per sec to wheel velocities...
        
        # global velocities for the simulated robot
        self.gvel_left_cmpersec = 0.0
        self.gvel_right_cmpersec = 0.0
        
        # global value to hold the robot's estimated (noisy) position
        self.gpos = ER2OdometricPosition()
	# here is the robot's actual odometry
	self.gposReal = ER2OdometricPosition()
        # pose and sensor history with timestamps
	self.fullHistory = []
        # each entry is (x,y,thr,(sensor tuple),timestamp)
        self.fullHistory.append( (0.0,0.0,0.0,(),0.0,(0.0,0.0,0.0)) )
        # update will update this full history...

        # global timestamp: last position-updated time
        # we assume the velocity is constant between updates of this variable
        # this time.clock() needs to be called to start the clock...
        self.lastUpdatedTime = time.clock() # initialize to about zero
        # get the map - must be done before the call to self.InitializeRobot,
        # which calls update, which needs the map...
        self.map = sim.Map( mapfile )
        # initialize robot with port...
        self.PORT = PORT
        self.InitializeRobot(PORT)

    def motionErrorModelVelocities(self):
	# returns a pair of values: the "actual" right and left
	# wheel velocities in cm/sec

	# returning self.gvel_left_cmpersec and self.gvel_right_cmpersec
	# means there is no error at all

	return  (self.gvel_left_cmpersec  + random.gauss(0.0,1.0), \
                 self.gvel_right_cmpersec + random.gauss(0.0,1.0))

	# a real lean!
	#return  (self.gvel_left_cmpersec + 1.0, self.gvel_right_cmpersec - 1.0)


    def update(self, deltaTimeSeconds=None, updateOnlyPose=False):
        # we need to move everything forward in time
        # we will assume now that this is being called fast enough so
        # that all relevent events (crashes, etc.) are being caught
        # THIS MAY ALWAYS HAVE TO BE ASSUMED... !!

        # it does seem a good idea to be able to pass in an 
        # artificial deltaTimeSeconds, however. 
        # this will allow for realistic, non-real-time simulations
        # if lots of computation is required...

        # what's the "right" way to handle artificial and real time
        # together? (probably just not to do it)

        # we'll worry about turning sensors on/off later...

        # if there is no artificial time, we use time.clock()
        if deltaTimeSeconds == None:
            newtime = time.clock()
            deltaTimeSeconds = newtime - self.lastUpdatedTime
        else:
            newtime = self.lastUpdatedTime + deltaTimeSeconds

        # OK, we know the time delta - now, how about the real vs.
	# the noisy robot location?

	# Well, let's assume that the estimated, noisy robot position
	# is the location obtained by integrating the commanded
	# velocities over time  (this, indeed, will be noisy)

        # update the [potential!] pose changes
        left_wheel_cm = self.gvel_left_cmpersec * deltaTimeSeconds
        right_wheel_cm = self.gvel_right_cmpersec * deltaTimeSeconds
        
        # next, update the current odometer based on those motions
        self.gpos.diffDriveKinUpdate(left_wheel_cm, right_wheel_cm)

	# Now, what about the REAL robot position - that is going 
	# to be some other location, based on nearby wheel velocities
	# we can imagine that slippage etc. can be modeled as
	# simply a (perhaps small) change in the commanded velocity
	# to each wheel. For the moment, we'll add a Gaussian-sampled
	# delta to each wheel independently, with the distribution
	# width proportional to wheel velocity
	# This way 0 velocity is really 0 velocity, which is fair.
        actual_left_cmpersec, actual_right_cmpersec = self.motionErrorModelVelocities()
        left_wheel_cmReal = actual_left_cmpersec * deltaTimeSeconds
        right_wheel_cmReal = actual_right_cmpersec * deltaTimeSeconds
        
        # next, update the current odometer based on those motions
        self.gposReal.diffDriveKinUpdate(left_wheel_cmReal, right_wheel_cmReal)

        # next, check for collisions...
        C = self.map.findER1collisions(self.gposReal.x, \
			self.gposReal.y, self.gposReal.thr)

        # if collision, we assume it did not move from the last position
        if C:
            lastState = self.fullHistory[-1]
            x = lastState[0] # should we reset these?
            y = lastState[1] # the odometric (not "real") info?
            thr = lastState[2]
	    realPos = lastState[5]
	    #self.gpos.set(x,y,thr)
            self.gposReal.set(realPos[0], realPos[1], realPos[2])
            print 'Collision!'
            stalled = True
        else:
            stalled = False

        # next, update active sensors
        # no notion of "active" quite yet...
        # note that this uses the updated self.gpos, which is
        # good, because we want to return the sensor values
        # of the previous pose if, indeed, the robot is stuck
        if updateOnlyPose == False:
            R = self.map.findER1ranges(self.gposReal.x, \
			    self.gposReal.y, self.gposReal.thr)
        else:
            R = []

        # put together pose and sensor data with timestamp
        # add to history
        x = self.gpos.x
        y = self.gpos.y
        thr = self.gpos.thr
        sensorTuple = (R,stalled)
	realPos = (self.gposReal.x, self.gposReal.y, self.gposReal.thr)
        # newtime is the timestamp
        self.fullHistory.append( (x,y,thr,sensorTuple,newtime,realPos) )

        # regardless, we set the lastUpdatedTime
        self.lastUpdatedTime = newtime


    def clearHistory():
        if len(self.poseHistory) > 0:
            self.poseHistory = self.poseHistory[-1:]
        # otherwise, it's already empty
    
    def SetVelocities_cm_deg(self, cmpersec, degpersec):
        # need to invert the kinematics to get the wheel rotations
        # in cm per second
    
        # the translational component is easy
        left_trans_cm_sec = cmpersec
        right_trans_cm_sec = cmpersec
    
        # the rotational component is not too bad 
        # note that if the translational component is 0, this
        # allows for a maximum of about 60 degrees per second...
    
        # ok, I think this does not combine rotation and translation
        # in the most natural way
        DELTA = self.DELTA
        left_rot_cm_sec = - self.DELTA * (degpersec * math.pi / 180.0 )
        right_rot_cm_sec = self.DELTA * (degpersec * math.pi / 180.0 )
    
        #print 'cm, deg', cmpersec, ',', degpersec
        #print 'left,right trans', left_trans_cm_sec, ',', right_trans_cm_sec
    
        # instead, we should add a constant that makes the smaller of
        # these equal to zero - that way the slower wheel will travel
        # the desired translational amount, and the rotation will
        # be taken up by the faster wheel...
    
        # since they're of opposite sign, this will work
        #adjust = max(left_rot_cm_sec, right_rot_cm_sec)
        #left_rot_cm_sec += adjust
        #right_rot_cm_sec += adjust
    
        # add them up
        left_cm_sec = left_trans_cm_sec + left_rot_cm_sec
        right_cm_sec = right_trans_cm_sec + right_rot_cm_sec
        #print 'left,right total', left_cm_sec, ',', right_cm_sec
    
        # send off to robot
        self.SetWheelVelocities(left_cm_sec, right_cm_sec)
    
    def SetWheelVelocities(self, left_cm_sec, right_cm_sec):
        # we need to update the odometry because whenever the velocity changes,
        # so does the interpretation of the distance traveled by each wheel...
        self.update(updateOnlyPose=True)
        # note that this call to UpdateOdometry has to be done FIRST,
        # so that the last time step is interpreted correctly

        #global gpos, gvel_right_cmpersec, gvel_left_cmpersec
        # why change if there's no change to make?
        if self.gvel_right_cmpersec == right_cm_sec and self.gvel_left_cmpersec == left_cm_sec: return

        # we can comment out speed limits, if we want...
        # but they're here for now!
        maxSpeed = max(left_cm_sec, right_cm_sec, -left_cm_sec, -right_cm_sec)
    
        # much bigger speed limit here in simulation... (vs. 20 or so)
        SPEEDLIMIT = 2000.0  # centimeters per second
    
        if maxSpeed > SPEEDLIMIT:
            left_cm_sec *= SPEEDLIMIT/maxSpeed
            right_cm_sec *= SPEEDLIMIT/maxSpeed
    
        # otherwise, set the velocities and wait for next chance to update the odometry
        self.gvel_right_cmpersec = right_cm_sec
        self.gvel_left_cmpersec = left_cm_sec
    
    def Go(self, cmpersec, degpersec=0.0):
        self.SetVelocities_cm_deg(cmpersec,degpersec) 


    # keep the optional port number just for compatibility - it gets ignored...
    def InitializeRobot(self, PORT=0):
        # this is left over from the real robot...
        self.update(updateOnlyPose=True)
        #global gpos, gvel_left_cmpersec, gvel_left_cmpersec
        self.gpos.reset()
        self.gvel_right_cmpersec = 0.0
        self.gvel_left_cmpersec = 0.0
        # sometimes they were not 0 at this point for reasons unknown...
    
    def SetOdometry(self, x,y,thr):
        #global gpos, gtime
        self.gpos.set(x,y,thr)
        # reset global time (gtime) via call time.clock() so that the next update
        # is from this point in time and not some past point in time...
        self.gtime = time.clock()
        # note that in the real robot, this would have to
        # do a bit more in order to get the relative motor counts 
        # updated to the current time, as well...
    
    def GetOdometry(self):
        self.update(updateOnlyPose=True)
        return [self.gpos.x, self.gpos.y, self.gpos.thr]

    def GetAllDataFrom( self, xp, yp, thrp ):
        # returns data for item (particle) at the input pose
        # first, check for in bounds!
        numhits = self.map.numHits(xp,yp,thrp)
        R = self.map.findER1ranges(xp, yp, thrp)
        return (R,numhits)

    def GetAllData(self):
        # update sensors, as well!
        self.update(updateOnlyPose=False)
        # return the last tuple...
        return self.fullHistory[-1]
    
    # "Where" for the lazy...
    def Wh(self):
        return self.GetOdometry()
        
    def PowerOff(self):
        # this will update the odometry one last time...
        # it will also briefly turn the power on (if it wasn't already)
        self.SetWheelVelocities(0.0,0.0)
    
    def CloseRobot(self):
        # want to keep APIs the same for sim and real robots
        self.SetWheelVelocities(0.0,0.0)
        pass


