# poseEstimator.py
#
# routines for 2d pose estimation from 1d visual features

# should import the Numeric routines we need...
import math
import scipy
import LinearAlgebra
import Numeric


def poseEstimate( WF, imf, SP = [] ):  # SP is not used now
    """ finds the camera pose that gives rise to
        the image features in imf
        corresponding to the world coordinates in WF

        so len(WF) should == len(imf)

        ( cx, cy, thd ) is returned where

          cx  is the camera's center in world coordinates
          cy  is the camera's center in world coordinates
          thd is the angle of the camera's optical axis in world coordinates (in degrees)
    """
    #
    # we create our big constraint matrix named M
    # the number of features is NF
    if len(WF) != len(imf):
        print "The numbers of image and world features are not equal!"
        print "imf is", imf
        print "WF is", WF
        return (0.0,0.0,0.0)

    #
    # M has 2*NF rows and NF+6 cols
    # 
    M = []
    NF = len(imf)  # number of features
    onerow = [0] * (NF+6)
    for ri in range(2*NF):
        M.append(onerow[:])  # slicing copies the row, otherwise crazy aliasing occurs

    # fill the constraint matrix M appropriately
    # two rows at a time
    for featurenum in range(NF):
        # row 1
        ri = 2*featurenum + 0  # row index
        M[ri][0] = WF[featurenum][0]
        M[ri][1] = 0.0
        M[ri][2] = WF[featurenum][1]
        M[ri][3] = 0.0
        M[ri][4] = 1.0
        M[ri][5] = 0.0
        featurecol = 6+featurenum

        # ci is the column index
        for ci in range(6,NF+6):
            if ci == featurecol:  M[ri][ci] = -imf[featurenum]
            else:                 M[ri][ci] = 0.0

        # row 2
        ri = 2*featurenum + 1  # again, the row index (now the next row)
        M[ri][0] = 0.0
        M[ri][1] = WF[featurenum][0]
        M[ri][2] = 0.0
        M[ri][3] = WF[featurenum][1]
        M[ri][4] = 0.0
        M[ri][5] = 1.0
        featurecol = 6+featurenum

        # again, ci is the column index
        for ci in range(6,NF+6):
            if ci == featurecol:  M[ri][ci] = -1.0
            else:                 M[ri][ci] =  0.0


    # M is now ready, and we create an equivalent array for Numaric
    # A is the array with M's contents - needed for LinearAlgebra
    A = Numeric.array(M)
    #print "A is", A

    # find the singular value decomposition
    # the svd has an optional argument for which we're using
    # the default ... not sure if there is an impact
    U, S, Vt = LinearAlgebra.singular_value_decomposition(A)

    #print "U is", U
    print "S is", S  # not really diagonal, but a vector of eigenvalues
    #print "Vt is", Vt
    
    # should be all zeros
    # check = A - Numeric.matrixmultiply(U*S, Vt) 
    # print "check is", check

    # this should be in math or somewhere, but I can't find it...
    def sgn(x):
        if x >= 0: return 1
        else: return -1

    # we check the original eigenvector (ev) for whether the scale factors are
    # positive or negative and we multiply the whole thing
    # by -1 if they're negative
    # this is imposing the constraint that features (or
    # at least the majority of them) should be in front of the
    # camera - indeed, something should happen if there
    # is a mix of front/back features (perhaps throwing
    # out this correspondence!)
    ev = Vt[-1]
    directions = [ sgn(x) for x in ev[6:] ]
    majority = sum(directions)
    # print "majority is", majority

    if majority < 0:
        ev = map( lambda x: -1.0*x, ev )

    (A,B,C,D,E,F) = ev[:6]
    # print "Matrix before scaling is\n", ("%6.3f  %6.3f  %6.3f\n" % (A, C, E)), ("%6.3f  %6.3f  %6.3f\n" % (B, D, F))

    # in other words,
    # A = ev[0], B = ev[1], C = ev[2], D = ev[3], E = ev[4], F = ev[5]
    #
    # and the recovered camera matrix is
    #
    # [ A C E ]
    # [ B D F ]
    #
    # we scale the whole thing to make the left-hand 2x2 submatrix
    # a rotation matrix...
    #
    # if the determinant is negative, however, this indicates that there has been
    # a flip (of the X axis) around the Y axis. This is because in the
    # canonical camera the axes' handedness is reversed (how sure am I of this?)
    # at least it seems to work...
    #
    det = (A*D - B*C)
    if det < 0:
        ev[0] = -1.0*ev[0]
        ev[2] = -1.0*ev[2]
        ev[4] = -1.0*ev[4]

    # now, the determinant should be positive
    # scale the whole thing to make it 1.0
    detscale = 1.0 / math.sqrt(math.fabs(det))
    # multiply by this
    # and check for angle...
    ev = map( lambda x: detscale*x, ev )
    (A,B,C,D,E,F) = ev[:6]

    #print "Final matrix is\n", ("%8.3f  %8.3f  %8.3f\n" % (A, C, E)), ("%8.3f  %8.3f  %8.3f\n" % (B, D, F))

    # seems nice...
    def applyCamera( (WX,WY), (A,B,C,D,E,F) ):
        cx = A*WX + C*WY + E
        cy = B*WX + D*WY + F
        return (cx,cy)

    #print "Final matrix is\n", ("%8.3f  %8.3f  %8.3f\n" % (A, C, E)), ("%8.3f  %8.3f  %8.3f\n" % (B, D, F))

    # get the angle so far...
    thr = math.atan2(C,A)

    # OK, now we do not yet have the camera's location and angle wrt world coordinates
    # this is because the camera matrix transforms world coordinates
    # into canonical camera coordinates, which are characterized by
    #   having the distance perpendicular to the optical axis in the first component
    #   having the distance along the optical axis in the second component
    #
    # notice that these canonical camera coordinates correspond to having
    # the optical axis pointing in the positive Y direction, but this is
    # 90 degrees offset from 0 degrees thd in the world coordinate system
    # with maybe the x-axis flipping around the y-axis, as well (as mentioned above)
    # 
    # first, we undo the rotation that the camera matrix provides,
    #
    undoangle = -thr
    # we negate these offsets for the same reason
    cx =  -1.0* (  math.cos(undoangle)*E + math.sin(undoangle)*F )
    cy =  -1.0* ( -math.sin(undoangle)*E + math.cos(undoangle)*F )
    # the angle needs to be adjusted for the 90 degree (pi/2) offset
    # between canonical camera coordinates and world coordinates
    angler = (thr+math.pi/2.0)
    angled = angler * 180.0/math.pi
    #print "camera center is x =%8.3f  y =%8.3f  and angle = %6.3f" % (cx,cy,angle)

    return ( cx, cy, angler )



def cameraTransform( WPt, cx, cy, cthr ):
    X, Y = WPt
    costhr = math.cos(cthr)
    sinthr = math.sin(cthr)
    # rotate
    distAlongOpAx = (X-cx)*costhr + (Y-cy)*sinthr
    distPerpOpAx = -(X-cx)*sinthr + (Y-cy)*costhr
    # print "  X,Y,Along,Perp:", X, Y, distAlongOpAx, distPerpOpAx
    return (1.0 * distPerpOpAx / distAlongOpAx)
    #if distAlongOpAx != 0.0:
    #    return (1.0 * distPerpOpAx / distAlongOpAx)
    #else:
    #    return 1.0e12  # something big

def cameraResiduals( p, X ):
    cx, cy, cthr = p
    err = [ float( x[0] - cameraTransform( x[1], cx, cy, cthr ) ) \
            for x in X ]
    while len(err) < 3:  err[:0] = [0] # needs to be at least len 3, it seems
    return err

def poseEstimate3( WF, imf, SP ):
    """ three-point pose estimate 
        WF = list of world features (2 dimensions each)
	imf = list of image features (1 dimension each)
        SP = starting pose (list of x, y, thr)
	focalLength = 1.0 for now
    """
    #print "imf is", imf
    #print "WF is", WF
    #print "SP is", SP
    # we need to create X, which is a bunch of
    # [ (imx, (WPtx, WPty)), ... ]
    X = []
    for i in range(0,len(WF)):
	X.append( (imf[i], WF[i]) )

    plsq = scipy.optimize.leastsq( cameraResiduals, SP, args=(X), full_output=0 )

    return plsq[0]

def poseEstimate12( WF, imf, SP ):
    """ three-point pose estimate 
        WF = list of world features (2 dimensions each)
	imf = list of image features (1 dimension each)
        SP = starting pose (list of x, y, thr)
	focalLength = 1.0 for now
    """
    #print "imf is", imf
    #print "WF is", WF
    #print "SP is", SP
    # we need to create X, which is a bunch of
    # [ (imx, (WPtx, WPty)), ... ]
    X = []
    for i in range(0,len(WF)):
	X.append( (imf[i], WF[i]) )

    output = scipy.optimize.fsolve( cameraResiduals, SP, args=(X), full_output=0 )

    #print "output is", output
    return output


if __name__ == "__main__":
    """ test some things out... """

    gcx = 0.0  # camera's center x (world coordinates)
    gcy = 0.0  # camera's center y (ditto)
    gthd = 0.0 # angle of camera's optical axis (world coords)
    focallength = 1.0  # in world coordinates

    import sys
    import asyncore
    import asyncClient
    import time

    # see if the user has overridden any of these values...
    L = len(sys.argv)
    if L > 1: gcx = float(sys.argv[1])
    if L > 2: gcy = float(sys.argv[2])
    if L > 3: gthd = float(sys.argv[3])
    if L > 4: focallength = float(sys.argv[4])

    # world feature list: WF
    # WF = [ (30,10), (30,-10), (10,15), (10,-15), (50,20), (50,-20) ]
    # WF = [ (30, 0), (30,10), (30,-10) ]
    WF = []
    for i in range(0,3):
        WF.append( (20 + 20*scipy.rand(), 0 + (-20 + 40*scipy.rand())) )

    #WF = [ (20,0), (20,10), (20,-10) ]
    WF = [ (20,-10), (20,10) ]

    # the corresponding "camera" feature list (in camera coordinates): CF
    # nice to have this intermediate step...
    imf = [ cameraTransform( WPt, gcx, gcy, gthd ) for WPt in WF ]

    NOISE = 0.00
    imf_noisy = [ imx + NOISE*scipy.randn() for imx in imf ]

    #print "WF is", WF
    #print "imf is", imf

    #SP = [ 1.5, 1.5, 0.2 ]
    SP = [ 15.5, -14.5, 1.0 ]

    if len(WF) > 0: # this top one seems better than the second one
        cx, cy, cthr = poseEstimate3( WF, imf_noisy, SP )
        # in that it's closer to the initial point when underconstrained
    else:
        cx, cy, cthr = poseEstimate12( WF, imf_noisy, SP )

    print "cx is", cx
    print "cy is", cy
    print "cthd is", 180.0*cthr/math.pi


    #import sensorSimulator as sim
    #m = sim.Map('.\\map.txt')
    #print "m is", m
    # test the image stuff
    #result = sim.getExactImageFeatures(50,50,225)
    #result = sim.getExactImageFeatures(0,0,0.0)
    #print "exact features are", result
    #result = m.getExactImage(50,50,225)
    #result = m.getExactImage(0,0,0.0)
    #print "exact image is", result

            #imf = []   # image features
            #WF = []    # world features
            #for k in exactimf.keys():     # the key is each image feature
            #    imf.append( k )
            #    WF.append( exactimf[k] )  # the value is the world feature
                #print "imf is", imf
                #print "WF is", WF
            # 
            # now, we TRY to get the correct pose
            # if there is any kind of singularity, this will fail
            # STILL NEED: singularity-handling...
            #cx, cy, angle = poseEstimate( WF, imf )
            #print "*** resulting camera: ", cx, cy, angle
            #
            # send the appropriate string to the visualizer to display the camera
            #camera = "c_%f_%f_%f_%f_%f_%s" % (cx,cy,thd,10,10,"orange")




